////////////////////////////////////////////////////////////////////////////////
/// \class HostModel HostModel.h
///
/// \version 1.00
/// \date 18-07-2006
/// \author Andrew Smith <andrew.smith@freenet.co.uk>
///
/// \par Copyright:
/// Copyright (C) 2005-2005 Trolltech AS. All rights reserved.
///
/// \brief Overridden directory model to use same file size format.
///
/// QDirModel currently outputs files sizes as int(size/1000) KB whereas the
/// rest of the application uses double(size/1024.0) KB, as does the OS.
/// This overridded directory model simply uses the QDirModel data but
/// calculates it's own file sizes using the same display/calculation function
/// as the rest of the application.
/// \sa QDirModel.
///
////////////////////////////////////////////////////////////////////////////////
#ifndef HOSTMODEL_H
#define HOSTMODEL_H

#include <QDirModel>

#include "HumaxGlobals.h"

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
class HostModel : public QDirModel
{
	/*! \brief Returns the data for the model item.
	 *
	 * Returns the data for the model item index with the given role.
	 * Reimplemented from QDirModel. Will return the contents of QDirModel
	 * with the exception of the size column, which is calculated using our
	 * own function.
	 * \param index	The index in the model that the data is for.
	 * \param role	The type of data needed.
	 * \return 		A variant containing the requested data.
	 */
	inline virtual QVariant data(const QModelIndex &index,
									int role = Qt::DisplayRole) const
	{
		if( role == Qt::DisplayRole &&
			index.column() == 1 &&
			!isDir(index) )
		{
			QFileInfo info = fileInfo(index);
			return QVariant(Humax::SizeAsString(info.size()));
		}
		return QDirModel::data(index, role);
	}
};

#endif // HOSTMODEL_H
////////////////////////////////////////////////////////////////////////////////
