////////////////////////////////////////////////////////////////////////////////
/// \namespace Humax 
//  \file HumaxGlobals.h
///
/// \version 1.00
/// \date 14-07-2006
/// \author Andrew Smith <andrew.smith@freenet.co.uk>
///
/// \par Copyright:
/// See COPYING file that comes with this distribution.
///
/// \brief Contains any global utility functions.
///
/// A namespace to contain any utility functions needed within this application.
///
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
///// CHANGE HISTORY                                                       /////
////////////////////////////////////////////////////////////////////////////////
/*****

Date:     Ver: Description:
-------- ---- ------------------------------------------------------------------
14-07-06 1.00 Initial release.

*****/
#ifndef HUMAXGLOBALS_H
#define HUMAXGLOBALS_H

#include <QString>

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
namespace Humax
{
	/*! \brief Converts the size to a string.
	 *
	 * Converts the 'size' in bytes to a string with the units shown,
	 * Eg bytes, KB, or MB.
	 * \param size			The size in bytes to be converted.
	 * \param fieldWidth	The width of the number part of the string.
	 * \param prec			The floating point precision to be used in the string.
	 * \return The size converted to a string.
	 */
	const QString SizeAsString(const qint64 &size,
								const int fieldWidth = 0,
								const int prec = 2);
};

#endif // HUMAXGLOBALS_H
////////////////////////////////////////////////////////////////////////////////
