////////////////////////////////////////////////////////////////////////////////
/// \class MainWindow MainWindow.h
///
/// \version 0.99a
/// \date 19-07-2006
/// \author Andrew Smith <andrew.smith@freenet.co.uk>
///
/// \par Copyright:
/// See COPYING file that comes with this distribution.
///
/// \brief \b ** 
///
/// --
///
////////////////////////////////////////////////////////////////////////////////
#ifndef MAINWINDOW_H
#define MAINWINDOW_H

#include "ui_MainWindow.h"

#include <QMainWindow>

#include "HumaxController.h"

class QHostProxyModel;
class HostModel;
class HumaxModel;

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
class MainWindow : public QMainWindow
{    
	Q_OBJECT
public:
	MainWindow(QWidget *parent = 0);

public slots:
	//File Menu.
	void on_mpActionConnect_triggered();
	void on_mpActionDisconnect_triggered();
	void on_mpActionFormat_triggered();
	void on_mpActionPreferences_triggered();

	//Tools->Host Menu.
	void on_mpActionHostNewFolder_triggered();
	void on_mpActionHostRename_triggered();
	void on_mpActionHostDelete_triggered();
	void on_mpActionHostRefresh_triggered();
	//Tools->Device Menu.
	void on_mpActionDeviceNewFolder_triggered();
	void on_mpActionDeviceRename_triggered();
	void on_mpActionDeviceDelete_triggered();
	void on_mpActionDeviceRefresh_triggered();
	//Tools Menu.
	void on_mpActionUpload_triggered();
	void on_mpActionDownload_triggered();

	//About Menu.
	void on_mpActionAbout_triggered();
	void on_mpActionWhatsThis_triggered();

	//Toolbar...
	void on_mpActionFilter_triggered(bool checked = false);
	// bool doInfo(usb_dev_handle * dh, bool fullInfo)
	void on_mpActionInformation_triggered();

	//Host Area...
	void on_mpHostDriveCombo_currentIndexChanged(int index);
	void on_mpHostParentButton_clicked();
	void on_mpHostDriveView_itemSelectionChanged();
	void on_mpHostDriveView_doubleClicked();

	//Device Area...
	void on_mpDeviceDriveCombo_currentIndexChanged(int index);
	void on_mpDeviceParentButton_clicked();
	void on_mpDeviceDriveView_itemSelectionChanged();
	// bool doChdir(usb_dev_handle * dh);
	void on_mpDeviceDriveView_doubleClicked();

	void transferError(QString error);

public:
	void Log(const QString &text);
	void Output(const QString &text);
	void Debug(const QString &text);
	void Warning(const QString &text);
	void Error(const QString &text);

	inline void ShowLogWindow(const bool &show)
		{ ui.mpOutputText->setVisible(show); }

	inline bool IsShowingLogWindow() const
		{ return ui.mpOutputText->isVisible(); }

	inline void SetTransferBlockSizes(const int &download, const int &primary, const int &recovery)
		{	mHumaxController.downloadBlockSize			= download;
			mHumaxController.uploadPrimaryBlockSize		= primary;
			mHumaxController.uploadRecoveryBlockSize	= recovery; }
	inline void SetLoggingFlags(const bool &output, const bool &errors, const bool &debug)
		{ mLogOutput = output; mLogErrors = errors; mLogDebug = debug; }
	inline void SetTraceFlags(const bool &console, const bool &logFile)
		{ mHumaxController.TraceConsole(console); mHumaxController.TraceLogFile(logFile); }

	inline int DownloadBlockSize() const { return mHumaxController.downloadBlockSize; }
	inline int UploadPrimaryBlockSize() const { return mHumaxController.uploadPrimaryBlockSize; }
	inline int UploadRecoveryBlockSize() const { return mHumaxController.uploadRecoveryBlockSize; }

	inline void ErrorRecovery(const bool &recover) { mHumaxController.ErrorRecovery(recover); }
	inline bool ErrorRecovery() const { return mHumaxController.ErrorRecovery(); }

	inline void ProcessPriority(const int &pri) { mHumaxController.processPriority = pri; }
	inline int ProcessPriority() const { return mHumaxController.processPriority; }

	inline bool LogOutput() const { return mLogOutput; }
	inline bool LogErrors() const { return mLogErrors; }
	inline bool LogDebug() const { return mLogDebug; }

	inline bool TraceConsole() const { return mHumaxController.TraceConsole(); }
	inline bool TraceLogFile() const { return mHumaxController.TraceLogFile(); }

protected:
	void closeEvent(QCloseEvent *event);

private:
	void Connections();

	void UpdateGui();

	// bool doList(usb_dev_handle * dh);
	bool UpdateDeviceList();

	void UpdateFilter();

	void readSettings();
	void writeSettings();

	bool HostDeleteFile(QString fileName);
	bool HostCopyFileList(const QString &basePath, const QString &hostDirPath,
							const QString &devPath, QStringList &downloadList,
							QStringList &mkDirList);

	bool DeviceDeleteFile(QString fileName);

	Ui::MainWindow ui;

	HumaxController mHumaxController;

	HostModel	*mpHostPathModel;
	HumaxModel	*mpHumaxModel;

	QHostProxyModel *mpHostProxyModel;

	bool mLogOutput;
	bool mLogErrors;
	bool mLogDebug;
};

#endif // MAINWINDOW_H
