////////////////////////////////////////////////////////////////////////////////
/// \class PrefsDialog PrefsDialog.h
///
/// \version 1.00
/// \date 14-07-2006
/// \author Andrew Smith <andrew.smith@freenet.co.uk>
///
/// \par Copyright:
/// See COPYING file that comes with this distribution.
///
/// \brief \b ** Media Controller Preferences Dialog Class.
///
/// Preferences dialog to display information about the current
/// Transfer and Logging settings.
///
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
///// CHANGE HISTORY                                                       /////
////////////////////////////////////////////////////////////////////////////////
/*****

Date:     Ver: Description:
-------- ---- ------------------------------------------------------------------
14-07-06 1.00 Initial release.

*****/
#ifndef PREFSDIALOG_H
#define PREFSDIALOG_H

#include "ui_PrefsDialog.h"

#include <QDialog>

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
class PrefsDialog : public QDialog
{    
	Q_OBJECT
public:
	/*! \brief Constructor.
	 *
	 * Construct and initialise the data in the Preferences Dialog Class.
	 */
	PrefsDialog(QWidget *parent = 0);

public slots:
	/*! \brief 
	 *
	 */
	virtual void accept();

	/*! \brief 
	 *
	 */
	void on_applyButton_clicked();

	/*! \brief 
	 *
	 * \param i
	 */
	void on_mpDownloadBlockSize_valueChanged(int i);

	/*! \brief 
	 *
	 * \param i
	 */
	void on_mpUploadPrimaryBlockSize_valueChanged(int i);

	/*! \brief 
	 *
	 * \param i
	 */
	void on_mpUploadRecoveryBlockSize_valueChanged(int i);

	/*! \brief 
	 *
	 * \param index
	 */
	void on_mpErrorRecovery_currentIndexChanged(int index);

	/*! \brief 
	 *
	 * \param index
	 */
	void on_mpProcessPriority_currentIndexChanged(int index);

	/*! \brief 
	 *
	 */
	void on_mpDefaults_clicked();

	/*! \brief 
	 *
	 * \param state
	 */
	void on_mpShowLogWindowToggle_stateChanged(int state);

	/*! \brief 
	 *
	 * \param state
	 */
	void on_mpShowOutputToggle_stateChanged(int state);

	/*! \brief 
	 *
	 * \param state
	 */
	void on_mpShowErrorsToggle_stateChanged(int state);

	/*! \brief 
	 *
	 * \param state
	 */
	void on_mpShowDebugToggle_stateChanged(int state);

	/*! \brief 
	 *
	 * \param state
	 */
	void on_mpTraceConsoleToggle_stateChanged(int state);

	/*! \brief 
	 *
	 * \param state
	 */
	void on_mpTraceLogFileToggle_stateChanged(int state);

private:
	/*! \brief 
	 *
	 */
	void UpdateSettings();

	/*! \brief Preferences Dialog widget information.
	 *
	 * Contains all of the information/widgets needed to created the dialog.
	 * This information is stored in PrefsDialog.ui and is auto-generated
	 * by Qt.
	 */
	Ui::PrefsDialog ui;

	/*! \brief 
	 *
	 */
	bool mDataChanged;
};

#endif // PREFSDIALOG_H
