////////////////////////////////////////////////////////////////////////////////
/// \class QDiskBar QDiskBar.h
///
/// \version 1.00
/// \date 04-07-2006
/// \author Andrew Smith <andrew.smith@freenet.co.uk>
///
/// \par Copyright:
/// See COPYING file that comes with this distribution.
///
/// \brief Disk Information Bar Widget Class.
///
/// A Qt Widget class to graphically display the space used on a disk partition.
///
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
///// CHANGE HISTORY                                                       /////
////////////////////////////////////////////////////////////////////////////////
/*****

Date:     Ver: Description:
-------- ---- ------------------------------------------------------------------
04-07-06 1.00 Initial release.

*****/
#ifndef QDISKBAR_H
#define QDISKBAR_H

#include <QFrame>

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
class QDiskBar : public QFrame
{    
	Q_OBJECT
public:
	/*! \brief Constructor.
	 *
	 * Construct and initialise the data in the Disk Bar Class.
	 */
	QDiskBar(QWidget *parent = 0);

	/*! \brief Destructor.
	 *
	 * Clean up any memory usage and destroy the class.
	 */
	inline virtual ~QDiskBar() {};

	/*! \brief Set the Disk space information.
	 *
	 * Set the disk space information to be displayed by the Disk Bar.
	 * \param free	The amount of free space on the disk.
	 * \param total	The total amount of space on the disk.
	 * \sa mFreeSpace and mTotalSpace.
	 */
	inline void setDiskSpace(const qint64 &free, const qint64 &total)
		{ mFreeSpace = free; mTotalSpace = total; }

protected:
	/*! \brief Paint the Disk Bar on the screen.
	 *
	 * Pain the Disk Bar on the dialog when a paint request event is made.
	 * \param event	The paint event request.
	 */
	virtual void paintEvent(QPaintEvent *event);

private:
	/*! \brief The amount of free space on the disk.
	 *
	 * Stores the amount of free space on the disk to be displayed by
	 * the disk bar.
	 * \sa setDiskSpace()
	 */
	qint64 mFreeSpace;

	/*! \brief The total amount of space on the disk.
	 *
	 * Stores the total amount of space on the disk to be displayed by
	 * the disk bar.
	 * \sa setDiskSpace()
	 */
	qint64 mTotalSpace;
};

#endif // QDISKBAR_H
////////////////////////////////////////////////////////////////////////////////
