////////////////////////////////////////////////////////////////////////////////
// \class AboutDialog AboutDialog.cpp
//
// \version 1.00
// \date 05-07-2006
// \author Andrew Smith <andrew.smith@freenet.co.uk>
//
// \par Copyright:
// See COPYING file that comes with this distribution.
//
// \brief Media Controller About Dialog Class.
//
// Basic about box dialog to display information about the Media Controller
// application including version and author information.
//
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
///// CHANGE HISTORY                                                       /////
////////////////////////////////////////////////////////////////////////////////
/*****

Date:     Ver: Description:
-------- ---- ------------------------------------------------------------------
05-07-06 1.00 Initial release.

*****/
#include "AboutDialog.h"

#include "humax.types.h"

////////////////////////////////////////////////////////////////////////////////
/*! \brief Constructor.
 *
 * Construct and initialise the data in the About Dialog Class.
 */
////////////////////////////////////////////////////////////////////////////////
AboutDialog::AboutDialog(QWidget *parent)
: QDialog(parent)
{
	ui.setupUi(this);

	QFont dlgFont = font();

	int pointSize = dlgFont.pointSize();
	if( pointSize == -1 )
	{
		pointSize = dlgFont.pixelSize();
	}

	QString smallFontSize = QString::number(pointSize);
	QString largeFontSize = QString::number(pointSize + 2);

	QString aboutText;
	aboutText +=
		"<html>"
			"<head>"
				"<meta name=\"qrichtext\" content=\"1\" />"
			"</head>"
			"<body style=\" white-space: pre-wrap; font-family:" + dlgFont.family() +
			"; font-size:" + smallFontSize + "pt; font-weight:400; font-style:normal; text-decoration:none;\">"
				"<p style=\" -qt-block-indent:0; text-indent:0px;\">"
					"Based on:"
				"</p>"
				"<p style=\" -qt-block-indent:0; text-indent:0px; font-size:" + largeFontSize + "pt; font-weight:600;\">"
					"<span>"
						+ tr(PROGRAM_NAME) + " " + tr(PROGRAM_VERSION) +
					"</span>"
					"<span style=\" font-size:" + smallFontSize + "pt; font-weight:400;\">"
						+ " (" + tr(VERSION_DATE) + ")" +
						"<br />"
					"</span>"
					"<span style=\" font-size:" + smallFontSize + "pt;\">"
						+ tr(COPYRIGHT) +
					"</span>"
				"</p>"
				"<br />"
				"<p style=\" -qt-block-indent:0; text-indent:0px;\">"
					"Using:"
				"</p>"
				"<p style=\" -qt-block-indent:0; text-indent:0px; font-size:" + largeFontSize + "pt; font-weight:600\">"
					"<span>"
						+ tr(GUI_PROGRAM_NAME) + " " + tr(GUI_PROGRAM_VERSION) +
					"</span>"
					"<span style=\" font-size:" + smallFontSize + "pt; font-weight:400;\">"
						+ " (" + tr(GUI_VERSION_DATE) + ")" +
						"<br />"
					"</span>"
					"<span style=\" font-size:" + smallFontSize + "pt;\">"
						+ tr(GUI_COPYRIGHT) +
					"</span>"
				"</p>"
			"</body>"
		"</html>";
	ui.mpAboutLabel->setText(aboutText);
}

////////////////////////////////////////////////////////////////////////////////
