////////////////////////////////////////////////////////////////////////////////
// \class HumaxController HumaxController.cpp
//
// \version 0.99a
// \date 04-07-2006
// \author Andrew Smith <andrew.smith@freenet.co.uk>
//
// \par Copyright:
// See COPYING file that comes with this distribution.
//
// \brief \b ** 
//
// --
//
////////////////////////////////////////////////////////////////////////////////
#include "HumaxController.h"

#include "FileReplaceDialog.h"
#include "HumaxGlobals.h"

#include <stdio.h>
#include <string.h>

#include <QFileInfo>

extern struct usb_device *mpHumaxDevice; // Handle to the USB device

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
HumaxController::HumaxController(QObject * parent)
: QThread(parent)
, HumaxInterface()
, mUploadFileList()
, mUploadDestDir()
, mDownloadBaseDir()
, mDownloadFileList()
, mDownloadDestDir()
, mPaused(false)
, mOverwriteReturnValue(0)
, downloadBlockSize(160)
, uploadPrimaryBlockSize(240)
, uploadRecoveryBlockSize(4)
, processPriority(3)
, mCancelled(false)
{
}
/*
HumaxController::~HumaxController()
{
	Disconnect();
}
*/

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
void HumaxController::run()
{
	mCancelled = false;
	bool overwriteAll = false;

	if( mDownloadFileList.count() > 0 && mUploadFileList.count() == 0 )
	{
		int id = 0;
		QString index;
		foreach(index, mDownloadFileList)
		{
			QString sourceFile = mDownloadBaseDir + index;

			Entry source;
			initEntry(&source, sourceFile.toStdString().data(), 0, 0);
			getFileInfo(&source);

			QString name = getFileName(source.path);

			emit transferStarted(id, name);

			QString targetFile = mDownloadDestDir + index;

			Entry target;
			initEntry(&target, targetFile.toStdString().data(), name.length(), 0);

			char path[MAX_PATH];
			/*int len =*/ getFilePath(target.path, target.nameLen, NULL, false, path);
			sprintf(target.path, "%s", path);
			target.nameLen = 0;

			if(setDeviceDirectory(&target) == -1)
			{
				deleteEntryTree(&target);
				emit transferError(QString("Failed to get device file info for '%1'").arg(targetFile));
				return;
			}

			// Check the contents of the first 1KB of data of the host file AND
			// the file on the device (by doing a short transfer over USB).
			// Return statuses are:
			// - FILE_NOT_FOUND : host file does not exist on the Humax
			// - FILE_EQUAL     : host file exists on the Humax and has the same file contents (well first 1KB of data anyway!)
			// - FILE_DIFFERENT : host file exists on the Humax but has different file contents
			int status = checkFile(&source, &target);
			if( status == FILE_DIFFERENT )
			{
				if( !overwriteAll )
				{
					Entry *file = target.firstChild;
					bool found = false;
					while( !found && file != NULL )
					{
						QString name2 = getFileName(file->path);
						found = (QString(name) == name2);
						if( !found )
							file = file->sibling;
					}

					mPaused = true;
					mOverwriteReturnValue = 0;
					emit transferOverwrite(name, Humax::SizeAsString(file->size),
						Humax::SizeAsString(source.size), (mDownloadFileList.count() == 1));

					while(mPaused)
					{
						msleep(100);
					}

					if( mOverwriteReturnValue == FileReplaceDialog::CANCEL )
					{
						//// OUTPUT SOME INFO!!!!
						emit transferComplete(id);
						mCancelled = true;
						break;
					}
					else if( mOverwriteReturnValue == FileReplaceDialog::NO )
					{
						//// OUTPUT SOME INFO!!!!
						emit transferComplete(id);
						deleteEntryTree(&target);
						continue;
					}
					else if( mOverwriteReturnValue == FileReplaceDialog::YESTOALL )
					{
						overwriteAll = true;
					}
				}
			}
			else if( status == FILE_EQUAL )
			{
				//// OUTPUT SOME INFO!!!!
				emit transferComplete(id);
				deleteEntryTree(&target);
				continue;
			}

			deleteEntryTree(&target);
			initEntry(&target, targetFile.toStdString().data(), name.length(), 0);

			xferFileHostToDevice(&source, &target);

			emit transferComplete(id);
			++id;
		}
		mDownloadFileList.clear();
		mDownloadDestDir = "";
	}

	if( mUploadFileList.count() > 0 && mDownloadFileList.count() == 0  )
	{
		int id = 0;
		Entry *index;
		foreach(index, mUploadFileList)
		{
			QString name = getFileName(index->path);

			emit transferStarted(id, name);

			QString targetFile = mUploadDestDir + name;

			Entry target;
			initEntry(&target, targetFile.toStdString().data(), 0, 0);
			target.nameLen = getValidHostFilePath(target.path, name.length());
			getFileInfo(&target);

			// Check the contents of the first 1KB of data of the host file AND
			// the file on the device (by doing a short transfer over USB).
			// Return statuses are:
			// - FILE_NOT_FOUND : host file does not exist on the Humax
			// - FILE_EQUAL     : host file exists on the Humax and has the same file contents (well first 1KB of data anyway!)
			// - FILE_DIFFERENT : host file exists on the Humax but has different file contents
			int status = checkFile(index, &target);
			if( status == FILE_DIFFERENT )
			{
				if( !overwriteAll )
				{
					mPaused = true;
					mOverwriteReturnValue = 0;
					emit transferOverwrite(name, Humax::SizeAsString(target.size),
						Humax::SizeAsString(index->size), (mUploadFileList.count() == 1));

					while(mPaused)
					{
						msleep(100);
					}

					if( mOverwriteReturnValue == FileReplaceDialog::CANCEL )
					{
						//// OUTPUT SOME INFO!!!!
						emit transferComplete(id);
						mCancelled = true;
						break;
					}
					else if( mOverwriteReturnValue == FileReplaceDialog::NO )
					{
						//// OUTPUT SOME INFO!!!!
						emit transferComplete(id);
						continue;
					}
					else if( mOverwriteReturnValue == FileReplaceDialog::YESTOALL )
					{
						overwriteAll = true;
					}

					QFile delFile(target.path);
					if( delFile.exists() )
					{
						delFile.remove();
					}
				}
				else
				{
					QFile delFile(target.path);
					if( delFile.exists() )
					{
						delFile.remove();
					}
				}
			}

			xferFileDeviceToHost(index, &target);

			emit transferComplete(id);
			++id;
		}
	}

	msleep(200);
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
void HumaxController::transferCancelled()
{
	mCancelled = true;
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
void HumaxController::transferOverwriteReturn(int retVal)
{
	mOverwriteReturnValue = retVal;
	mPaused = false;
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
bool HumaxController::Connect()
{
	mpUsbHandle = humaxConnect();
	return IsConnected();
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
bool HumaxController::IsConnected()
{
	return (mpUsbHandle != NULL);
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
bool HumaxController::Disconnect()
{
	if( IsConnected() )
	{
		bool retVal = humaxDisconnect();
		mpUsbHandle = NULL;
		return retVal;
	}
	return false;
}

////////////////////////////////////////////////////////////////////////////////
///// INITIALISATION FUNCTIONS                                             /////
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
/*! \brief Initialise.
 *
 * \return true - if initialisation succeeded, false - if initialisation failed.
 */
////////////////////////////////////////////////////////////////////////////////
bool HumaxController::Init()
{
	// Set initial device current directories and disk no
	initEntry(&deviceCurDir[0], "1:/", 0, 0);
	initEntry(&deviceCurDir[1], "2:/", 0, 0);
	deviceCurDiskNo = '2';

	// Return success
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// Do the change disk option (device disk 1 or 2)
// Inputs
// - dh           : USB device handle
// - deviceDiskNo : '1' for Image/MP3 disk, '2' for Video disk
//                  That's ASCII characters '1' and '2', not numerics 1 and 2!)
// Returns
// - true         : operation succeeded
// - false        : operation failed
////////////////////////////////////////////////////////////////////////////////
void HumaxController::ChangeDisk(char deviceDiskNo)
{
	// Set current device disk
	deviceCurDiskNo = deviceDiskNo;
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
bool HumaxController::Format()
{
	// Yes so set disk 1
	deviceCurDiskNo = '1';
	strcpy(deviceCurDir[0].path, "1:/");
///	strcpy_s(deviceCurDir[0].path, "1:/");
	if( !humaxSetDiskNo('1') )
	{
		emit transferError("Failed to change to disk 1!");
		return FALSE;
	}

	// Set root directory on device
	if( setDeviceDirectory(&deviceCurDir[0]) == -1 )
	{
		emit transferError("Failed to set root directory on disk 1!");
		return FALSE;
	}

	// Format disk 1
	if( !humaxFormatDiskOne() )
	{
		emit transferError("Failed to format disk partition 1!");
		return FALSE;
	}

	return TRUE;
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
bool HumaxController::doTransfersDeviceToHost(QList<Entry *> fileList, const QString &destDir)
{
	mUploadFileList	= fileList;

	mUploadDestDir	= destDir;
	if( mUploadDestDir.right(1) != "/" ) mUploadDestDir += "/";

	start(Priority(processPriority));

	return true;
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
bool HumaxController::doTransfersHostToDevice(const QString &basePath, const QStringList &fileList, const QString &destDir)
{
	mDownloadBaseDir	= basePath;

	mDownloadFileList	= fileList;

	mDownloadDestDir	= destDir;
	if( mDownloadDestDir.right(1) != "/" ) mDownloadDestDir += "/";

	start(Priority(processPriority));

	return true;
}

////////////////////////////////////////////////////////////////////////////////

// Create a new file or directory entry in memory
// Inputs
// - path  : Initial path
// - size  : Initial file size
// Returns
// - Entry : Newly allocated and initialised Entry in memory
Entry * HumaxController::createEntry(const char * path, int nameLen, qint64 size)
{
	Entry * e = new Entry;
	strcpy(e->path, path);
///	strcpy_s(e->path, path);
	e->nameLen = nameLen;
	e->size = size;
	e->isRecording = false;
	e->sibling = NULL;
	e->firstChild = NULL;
	return e;
}

// Initialises an existing file or directory Entry
// Inputs
// - e    : Entry to initialise, contents ignored on entry
// - path : Initial path
// - size : Initial file size
// Returns
// - e    : Newly initialised entry returned
void HumaxController::initEntry(Entry *e, const char *path, int nameLen, qint64 size)
{
	if(path == NULL)
		e->path[0] = 0;
	else
		strcpy(e->path, path);
///		strcpy_s(e->path, path);
	e->nameLen = nameLen;
	e->size = size;
	e->isRecording = false;
	e->index = -1;
	e->firstChild = NULL;
	e->sibling = NULL;
}

////////////////////////////////////////////////////////////////////////////////
/*! \brief Create a new Good Block structure.
 *
 * Create a new Good Block structure and links it to its previous and next
 * Block structures and return the new Good Block.
 * \param begByteNo	Start byte offset.
 * \param endByteNo	End byte offset.
 * \param curBlock	Current Block after which the new Block will be linked.
 * \param nextBlock	Next Block before which the new Block will be inserted.
 * \return The newly created Block.
 */
////////////////////////////////////////////////////////////////////////////////
Block *HumaxController::createBlock(qint64 begByteNo, qint64 endByteNo, Block *curBlock, Block *nextBlock)
{
	Block * newBlock = new Block;
	newBlock->begByteNo = begByteNo;
	newBlock->endByteNo = endByteNo;
	newBlock->nextBlock = nextBlock;
	if(curBlock != NULL)
		curBlock->nextBlock = newBlock;
	return newBlock;
}

////////////////////////////////////////////////////////////////////////////////
// Deletes a complete entry tree from memory preserving the top-level entry
// Inputs
// - dir : directory entry for which to do a hierarchical delete
// Returns
// - dir : directory entry with all sub-entries deleted
////////////////////////////////////////////////////////////////////////////////
void HumaxController::deleteEntryTree(Entry * dir)
{
	// If empty dir then just exit
	if(dir == NULL)
		return;

	// Loop for each entry in the current directory
	Entry * e = dir->firstChild;
	while(e != NULL)
	{
		// If current entry is directory then recurse
		if(isDirectory(e))
			deleteEntryTree(e);

		// Clear the entry and free memory
		Entry * sibling = e->sibling;
		free(e);

		// Do next entry
		e = sibling;
	}

	// Remove links to child and sibling entries
	dir->firstChild = NULL;
	dir->sibling = NULL;
}

// Deletes a block list from memory preserving the first entry
// Inputs
// - blocks : the first block to start deletion from
// Returns
// - blocks : first block returned with all subsequent linked blocks removed
void HumaxController::deleteBlockList(Block *blocks)
{
	// If no blocks then just exit
	if(blocks == NULL || blocks->nextBlock == NULL)
		return;

	// Loop for each entry
	Block * b = blocks->nextBlock;
	while(b != NULL)
	{
		// Free block from memory
		Block * nextBlock = b->nextBlock;
		free(b);
		b = nextBlock;
	}
}

//////////////////////////////////////////////////////////////////////////////////////
///// SINGLE FILE USB TRANSFER FUNCTIONS                                         /////
//////////////////////////////////////////////////////////////////////////////////////

// Transfer a single file from Humax to host over USB
// Inputs
// - dh         : USB device handle
// - sourceFile : source (Humax) file
// - targetFile : target (Host) file
// Returns
// - true       : if operation completed successully
// - false      : if file transfer failed with serious USB error or user aborted with the ESCAPE key
bool HumaxController::xferFileDeviceToHost(Entry * sourceFile, Entry * targetFile)
{
	// Setup transfer stats[xferNo][statNo]
	qint64 stats[11][6];
	for( int a = 0; a < 11; ++a )
		for( int b = 0; b < 6; ++b )
			stats[a][b] = 0;

	// Update source device file size info
	if(!getFileInfo(sourceFile))
	{
		emit transferError(QString("Failed to get source file size for '%1'!").arg(sourceFile->path));
		return false;
	}

	// If file currently being recorded on the Humax then engage chase-transfer mode
	bool chaseXferMode = sourceFile->isRecording;

	// Display file transfer details header
	QString sourceName = getFileName(sourceFile->path);
	QString targetName = getFileName(targetFile->path);

	bool nameChange = (sourceName != targetName);
	printf("  %sTransfer %s'%s'%s\n", (chaseXferMode ? "Chase-" : ""), (nameChange & !chaseXferMode ? " " : ""), sourceName.toStdString().data(), (chaseXferMode ? RECORDING_TAG : ""));
	if(nameChange)
		printf("  %sRename to '%s'\n", (chaseXferMode ? "     " : ""), targetName.toStdString().data());
	printf("  From Humax to Host %s Error Recovery, Size %.3f%s MB\n", (ErrorRecovery() ? "with" : "without"), (double)sourceFile->size / 1048576.0, (chaseXferMode ? "+" : ""));

	// If the Humax file is also present on the host then read the contents of that file first to check how much
	// of the file was previously transferred and to establish which parts of it contain good data
	char data[0x10000];
	Block goodBlocks;
	goodBlocks.begByteNo = -1;
	goodBlocks.endByteNo = -1;
	goodBlocks.nextBlock = NULL;
	bool doneComplete = false;
	int padFileBytes = 0;
	bool escape = false;

	QFile oFile;
	oFile.setFileName(targetFile->path);
	if( oFile.open(QIODevice::ReadOnly) )
	{
		// Get the current target file size
		getFileInfo(targetFile);
		padFileBytes = (256 - (int)(targetFile->size % 256)) % 256;
		targetFile->size = (targetFile->size + 255) / 256 * 256;

		// Get start time
		time_t startTime;
		time(&startTime);
		time_t nextTime = startTime + 1;

		// Initialise good data blocks list
		Block * block = &goodBlocks;
		qint64 curByteNo = 0;

		// Clear 64K data buffer and read the first block from the target file on the host
		memset(data, 0, 0x10000);
		qint64 dataSize = oFile.read(data, 0x10000);
		while(!escape && dataSize > 0)
		{
			// Check each 256-byte page boundary for signs that a data block wasn't previously transmitted
			// corrently over the USB connection
			for( int p = 0; p < dataSize; p = p + 256, curByteNo = curByteNo + 256 )
			{
				// First check for the presence of the "USBS" tag on the page boundary due to a previous transfer
				// made using the Humax Media eLinker software
				if(strncmp(&data[p], MISSING_TAG_ELINKER, strlen(MISSING_TAG_ELINKER)) == 0)
				{
					// Found missing data due to transfer previously made by the Humax Media eLinker program.
					// We need to mark the entire block (240 x 256 bytes = 60KB block) as bad. Media eLinker
					// always transmits files in 60KB blocks. So the way we determine the beginning and end of
					// the bad 60KB block is as follows:
					// - From the end of the page where the USBS error tag occurs, keep scanning forwards in the
					//   file until we hit a non-zero byte of data.
					// - The page where the first non-zero byte occurs is the start of the next good block of
					//   data so we can now mark the entire 240 x 256-byte blocks occurring before this page as
					//   bad.
					int a = 255;
					while(p + a < dataSize && data[p + a] == 0)
					{
						++a;
						if(a == 256)
						{
							a = 0;
							curByteNo = curByteNo + 256;
							p = p + 256;
							if(p == 0x10000)
							{
								// Ooop hit the end of our buffer so read more data in from the file
								dataSize = oFile.read(data, 0x10000);
								p = 0;
							}
						}
					}

					// Found start of next good data block so update stats
					if(block->begByteNo != -1 && curByteNo > 61441)
					{
						block->endByteNo = curByteNo - 61441;
						stats[0][BAD_BYTE_COUNT] = stats[0][BAD_BYTE_COUNT] + 61440;
					}
					else
						stats[0][BAD_BYTE_COUNT] = stats[0][BAD_BYTE_COUNT] + curByteNo;

					// Set pointers to start of last BAD block of data as these will get moved forward
					// to the next 256-byte page on the next iteration of the loop!
					curByteNo = curByteNo - 256;
					p = p - 256;
				}

				// Check if the ###MISSING### tag occurs at the 256-byte page start (the indicator of choice used
				// by this program)
				else if(strncmp(&data[p], MISSING_TAG_THIS, strlen(MISSING_TAG_THIS)) == 0)
				{
					// Found missing data due to transfer previously made by this program so update stats
					stats[0][BAD_BYTE_COUNT] = stats[0][BAD_BYTE_COUNT] + 256;
				}
				else
				{
					// Got a data page containing good data so update the Good Block data list entry.
					// If the previous page was bad then a new Good Block entry is created which excludes all the
					// previous Bad Block pages.
					if(block->begByteNo == -1)
					{
						block->begByteNo = curByteNo;
						block->endByteNo = curByteNo + 255;
					}
					else if((curByteNo - 1) != block->endByteNo)
						block = createBlock(curByteNo, curByteNo + 255, block, NULL);
					else
						block->endByteNo = curByteNo + 255;
				}
			}

			// Display the target file read status after each second elapses
			time_t curTime;
			time(&curTime);
			if(curTime >= nextTime)
			{
				stats[0][TOT_BYTE_COUNT] = curByteNo;
				stats[0][GOOD_BYTE_COUNT] = stats[0][TOT_BYTE_COUNT] - stats[0][BAD_BYTE_COUNT];
				stats[0][ELAPSED_SECS] = curTime - startTime;
				if(stats[0][ELAPSED_SECS] == 0)
					stats[0][ELAPSED_SECS] = 1;
				int remainSecs = (int)(((double)targetFile->size - stats[0][TOT_BYTE_COUNT]) / ((double)stats[0][TOT_BYTE_COUNT] / stats[0][ELAPSED_SECS]));

				emit uploadProgress(
					(int)((double)stats[0][TOT_BYTE_COUNT] / targetFile->size * 100.0),
					"File",
					((double)stats[0][GOOD_BYTE_COUNT] / 1048576.0),
					double(-1.0),
					((double)stats[0][BAD_BYTE_COUNT] / 1048576.0),
					((double)stats[0][TOT_BYTE_COUNT] / 1048576.0),
					double(-1.0),
					(int)stats[0][ELAPSED_SECS],
					remainSecs);

				nextTime = curTime + 1;
			}

			// Read the next 64KB of data from the target file
			memset(data, 0, 0x10000);
			dataSize = oFile.read(data, 0x10000);

			// Check if the ESCAPE key was pressed to abort the target file check
			escape = mCancelled;
		}

		// Close the target file for reading
		oFile.close();

		// Display the final file check stats
		time_t curTime;
		time(&curTime);
		stats[0][TOT_BYTE_COUNT] = curByteNo;
		stats[0][GOOD_BYTE_COUNT] = stats[0][TOT_BYTE_COUNT] - stats[0][BAD_BYTE_COUNT];
		stats[0][ELAPSED_SECS] = curTime - startTime;
		if(stats[0][ELAPSED_SECS] == 0)
			stats[0][ELAPSED_SECS] = 1;
		int remainSecs = (int)(((double)targetFile->size - stats[0][TOT_BYTE_COUNT]) / ((double)stats[0][TOT_BYTE_COUNT] / stats[0][ELAPSED_SECS]));

		emit uploadProgress(
			(int)((double)stats[0][TOT_BYTE_COUNT] / targetFile->size * 100.0),
			"File",
			((double)stats[0][GOOD_BYTE_COUNT] / 1048576.0),
			double(-1.0),
			((double)stats[0][BAD_BYTE_COUNT] / 1048576.0),
			((double)stats[0][TOT_BYTE_COUNT] / 1048576.0),
			double(-1.0),
			(int)stats[0][ELAPSED_SECS],
			remainSecs);

		// Check if the target file check completed (i.e. the source file on the Humax was already transferred previously
		// and successfully). If yes then no need to transfer the file again!
		doneComplete = isXferComplete(sourceFile->size + padFileBytes, &goodBlocks);
	}

	// Done the host file check so now time to do the USB transfer from Humax to host, but only if the file hasn't
	// already been transferred
	int xferNo = 0;
	bool ok = true;
	while(ok && !escape && !doneComplete && xferNo < (ErrorRecovery() ? 10 : 1))
	{
		// Increment transfer attempt counter
		// - Attempt 0 was the file check above
		// - Attempt 1 is the initial USB file transfer
		// - Attempts 2+ are the recovery transfers if there are still errors after the file check and initial transfer
		++xferNo;

		// Determine the total size of the file as stored on the Humax
		qint64 totXferBytes =
			(targetFile->size - padFileBytes == sourceFile->size || xferNo > 1 ? xferGetRecoveryByteCount(&goodBlocks) : sourceFile->size);

		// Open target host file for writing (or seek+write if already partially present on the host)
		oFile.setFileName(targetFile->path);
		if( !oFile.open((targetFile->size > 0 || xferNo > 1 ? QIODevice::ReadWrite : QIODevice::WriteOnly)) )
		{
			emit transferError(QString("Failed to open output file '%1'!").arg(targetFile->path));
			deleteBlockList(&goodBlocks);
			return false;
		}

		// Initiate the "M1" file transfer via USB
		int xferCBWTag = incCBWTag();
		if(!humaxXferBeginDevToHost(xferCBWTag, sourceFile->path))
		{
			emit transferError(QString("Failed to begin file transfer for '%1'!").arg(sourceFile->path));
			deleteBlockList(&goodBlocks);
			return false;
		}

		// Get the transfer start time
		time_t startTime;
		time(&startTime);
		time_t nextTime = startTime + 1;
		// Point to the current start of the Good Block list
		Block * block = &goodBlocks;

		// Now loop for each data block to read from USB
		qint64 curByteNo = 0;
		bool doneXfer = false;
		while(ok && !escape && !doneXfer && !doneComplete)
		{
			// Determin how much more data to read
			bool recoveryXfer = false;
			bool seekOnly = false;
			qint64 remainingByteCount = sourceFile->size - curByteNo;
			int thisByteCount = 0;

			// Check if we need to move to the next Good Block entry in our existing list
			if(block->nextBlock != NULL && curByteNo >= block->nextBlock->begByteNo)
				block = block->nextBlock;

			// Determine how many bytes to read via USB and the type of transfer we are doing for the current data block
			if(curByteNo <= block->endByteNo)
			{
				// We already have the current data marked as good data in the current Good Block entry so
				// just read the data via USB and throw it away (seek read). If we are near the end of the
				// current good block then reduce the USB read size for this block only so that we seek up
				// to the end of the current good block.
				// The "Humax-Host Xfer Primary Block Size" determines the maximum read size for initial and
				// seek reads.
				if(curByteNo + (uploadPrimaryBlockSize * 256) - 1 <= block->endByteNo)
					thisByteCount = uploadPrimaryBlockSize * 256;
				else
					thisByteCount = (int)(block->endByteNo - curByteNo + 1);
				seekOnly = true;
			}
			else if(block->nextBlock != NULL && curByteNo < block->nextBlock->begByteNo)
			{
				// Aha, we've hit a bad block of data so now we read this data ideally in 1KB data sizes in an
				// attempt to guarantee that we get this data (as determined by the "Humax-Host Xfer Recovery Block Size"
				// parameter. In tests, 1KB blocks were much slower but seemed to always get through without error.
				// If the remainder of the bad block is less than 1KB then adjusts the read so that we hit the end
				// of the bad block.
				// This transfer is a recovery read.
				if(curByteNo + (uploadRecoveryBlockSize * 256) - 1 < block->nextBlock->begByteNo)
					thisByteCount = uploadRecoveryBlockSize * 256;
				else
					thisByteCount = (int)(block->nextBlock->begByteNo - curByteNo);
				recoveryXfer = true;
			}
			else
			{
				// Just doing a primary read as we haven't so far read any of the current block so set read size to
				// the "Humax-Host Xfer Primary Block Size".
				thisByteCount = uploadPrimaryBlockSize * 256;
			}

			// If we're at the end of the file transfer then adjust the current read size so that we hit the end of
			// the file (otherwise the read will fail and the Humax might hang!)
			if(thisByteCount > remainingByteCount)
				thisByteCount = (int)remainingByteCount;

			// Read the next data block via USB using a data block size as determined above. cbRet determines the
			// amount of data we actually received
			int cbRet = humaxXferNextDevToHost(xferCBWTag, data, thisByteCount);

			// Check if we are just seek reading to skip over data or reading useful data not already acquired
			if(seekOnly)
			{
				// Nope, just seek reading so update good byte count and ignore what we read (even if it got corrupted!)
				// because we already have this data
				stats[xferNo][GOOD_BYTE_COUNT] += thisByteCount;
			}
			else
			{
				// We are doing a primary or recovery transfer so check how much data we actually received and
				// compare it to what we asked for. If what we got was less than what we asked for then the
				// transfer got corrupted for that block (possible due to USB driver issues) so we need to mark
				// the entire block as bad
				if(cbRet != thisByteCount)
				{
					// We have missing data from the block so invalidate the entire block and update the bad data count
					cbRet = 0;
					stats[xferNo][BAD_BYTE_COUNT] += thisByteCount;

					// Clear the entire data block
					memset(data, 0, thisByteCount);

					// Insert the ###MISSING### tag at start of each page of the bad block
					for(int p = 0; p < thisByteCount; p = p + 256)
						memcpy(&data[p], MISSING_TAG_THIS, strlen(MISSING_TAG_THIS));
				}
				else if(recoveryXfer)
				{
					// We were doing a recovery transfer which returned a good data block so update the
					// recovery good byte count
					stats[xferNo][RECOV_BYTE_COUNT] += thisByteCount;
				}
				else
				{
					// We were doing a primary or seek transfer so just update the good byte count
					stats[xferNo][GOOD_BYTE_COUNT] += thisByteCount;
				}

				// Update good block data if we have some
				if(cbRet != 0)
				{
					// Check where we are in the Good Block list
					if(block->begByteNo == -1)
					{
						// This is the first good data block so update begin and end byte pointers
						block->begByteNo = curByteNo;
						block->endByteNo = curByteNo + thisByteCount - 1;
					}
					else if(curByteNo > block->endByteNo)
					{
						// We are past the end of the current good data block
						if((curByteNo - 1) == block->endByteNo)
						{
							// The current USB read continues on from the end of the current Good Block so just
							// extend the end byte pointer of the current Good Block
							block->endByteNo = curByteNo + thisByteCount - 1;
						}
						else if((curByteNo - 1) > block->endByteNo)
						{
							// There's still some bad data from a previous USB read as we're in the middle of a
							// Bad Block so create a new Good Block entry and insert it into the current position
							// in the Good Block list
							block = createBlock(curByteNo, curByteNo + thisByteCount - 1, block, block->nextBlock);
						}
					}
				}

				// Read to write data to the host target file so seek the correct position in file
				if( !oFile.seek(curByteNo) )
				{
					emit transferError(tr("Seek failed!"));
				}

				// Write the data block to the target host file
				if( oFile.write(data, thisByteCount) != thisByteCount )
				{
					emit transferError(QString("Failed to write data to local file '%1'!").arg(targetFile->path));
					ok = false;
					continue;
				}
			}

			// Update current byte position counter
			curByteNo += thisByteCount;

			// Update flags to indicate the end of the current transfer number, and the end of the complete transfer
			doneXfer = curByteNo >= sourceFile->size;
			doneComplete = !sourceFile->isRecording && isXferComplete(sourceFile->size, &goodBlocks);

			// Check if the Humax file is still recording content (chase-transfer)
			if(doneXfer && sourceFile->isRecording)
			{
				// File still recording so chase-transfer the file as it expands
				doneXfer = false;

				// Read (via USB) the new file size and recording state info for the source Humax file
				getFileInfo(sourceFile);
				totXferBytes = sourceFile->size;

				// If less than 5MB from the current end of the Humax file then sleep for 10 seconds before resuming
				while(!escape && sourceFile->isRecording && (curByteNo + 5242880) >= totXferBytes)
				{
					// Sleep for 20 x 0.5 seconds, checking if the ESCAPE key pressed by the user to abort the transfer)
					// before resuming the transfer
					for(int a = 0; !escape && a < 20; a++)
					{
						// Check if ESCAPE key pressed by user
						escape = mCancelled;

						// Sleep for half a second and update chase-pause indicator
						printf("    CHASE%s\x0d", (a % 2 == 0 ? "+" : " "));
						msleep(500);
					}

					// Read (via USB) the new file size and recording state info of the source Humax file
					getFileInfo(sourceFile);
					totXferBytes = sourceFile->size;
				}

				// Get current time and next stats update time
				time_t curTime;
				time(&curTime);
				nextTime = curTime + 1;
			}

			// Dsiplay transfer stats to the user after each second elapses
			time_t curTime;
			time(&curTime);

			if(curTime >= nextTime)
			{
				// 1 second elapsed after last update so display stats to the user
				QString taskName;
				if( chaseXferMode && xferNo == 1 )
					taskName = "Chase";
				else if( xferNo == 1 )
					taskName = "Transfer";
				else
					taskName = QString("Recovery %1").arg(xferNo - 1);

				stats[xferNo][TOT_BYTE_COUNT] = curByteNo;
				stats[xferNo][RECOV_BYTE_COUNT] = stats[xferNo][TOT_BYTE_COUNT] - stats[xferNo][GOOD_BYTE_COUNT] - stats[xferNo][BAD_BYTE_COUNT];
				stats[xferNo][ELAPSED_SECS] = curTime - startTime;
				if(stats[xferNo][ELAPSED_SECS] == 0)
					stats[xferNo][ELAPSED_SECS] = 1;
				stats[xferNo][RATE_BYTES] = (int)(stats[xferNo][TOT_BYTE_COUNT] / stats[xferNo][ELAPSED_SECS]);
				int remainSecs = (int)(((double)totXferBytes - stats[xferNo][TOT_BYTE_COUNT]) / ((double)stats[xferNo][TOT_BYTE_COUNT] / stats[xferNo][ELAPSED_SECS]));

				emit uploadProgress(
					(int)((double)stats[xferNo][TOT_BYTE_COUNT] / totXferBytes * 100.0),
					taskName,
					((double)stats[xferNo][GOOD_BYTE_COUNT] / 1048576.0),
					((double)stats[xferNo][RECOV_BYTE_COUNT] / 1048576.0),
					((double)stats[xferNo][BAD_BYTE_COUNT] / 1048576.0),
					((double)stats[xferNo][TOT_BYTE_COUNT] / 1048576.0),
					((double)stats[xferNo][RATE_BYTES] / 1024.0),
					(int)stats[xferNo][ELAPSED_SECS],
					remainSecs);

				nextTime = curTime + 1;
			}

			// Check if ESCAPE key pressed by the user
			if(!escape)
				escape = mCancelled;
		}

		// End the file transfer via USB
		if( !humaxXferEnd(xferCBWTag, sourceFile->path) )
		{
			emit transferError(QString("Failed to end file transfer for '%1'!").arg(sourceFile->path));
			oFile.close();
			deleteBlockList(&goodBlocks);
			return false;
		}

		// Close the target host file
		oFile.close();

		// Display final transfer stats and status to the user
		time_t curTime;
		time(&curTime);

		QString taskName;
		if( chaseXferMode && xferNo == 1 )
			taskName = "Chase";
		else if( xferNo == 1 )
			taskName = "Transfer";
		else
			taskName = QString("Recovery %1").arg(xferNo - 1);

		stats[xferNo][TOT_BYTE_COUNT] = curByteNo;
		stats[xferNo][RECOV_BYTE_COUNT] = stats[xferNo][TOT_BYTE_COUNT] - stats[xferNo][GOOD_BYTE_COUNT] - stats[xferNo][BAD_BYTE_COUNT];
		stats[xferNo][ELAPSED_SECS] = curTime - startTime;
		if(stats[xferNo][ELAPSED_SECS] == 0)
			stats[xferNo][ELAPSED_SECS] = 1;
		stats[xferNo][RATE_BYTES] = (int)(stats[xferNo][TOT_BYTE_COUNT] / stats[xferNo][ELAPSED_SECS]);
		int remainSecs = (int)(((double)totXferBytes - stats[xferNo][TOT_BYTE_COUNT]) / ((double)stats[xferNo][TOT_BYTE_COUNT] / stats[xferNo][ELAPSED_SECS]));

		emit uploadProgress(
			(int)((double)stats[xferNo][TOT_BYTE_COUNT] / totXferBytes * 100.0),
			taskName,
			((double)stats[xferNo][GOOD_BYTE_COUNT] / 1048576.0),
			((double)stats[xferNo][RECOV_BYTE_COUNT] / 1048576.0),
			((double)stats[xferNo][BAD_BYTE_COUNT] / 1048576.0),
			((double)stats[xferNo][TOT_BYTE_COUNT] / 1048576.0),
			((double)stats[xferNo][RATE_BYTES] / 1024.0),
			(int)stats[xferNo][ELAPSED_SECS],
			remainSecs);
	}

	// Calculate and display the summary stats to the user
	double pct = ((double)(xferGetGoodByteCount(&goodBlocks) - padFileBytes) / sourceFile->size * 100.0);
	if((targetFile->size == 0 && xferNo > 1) || (targetFile->size > 0 && xferNo > 0))
	{
		// Calculate summary stats from all transfers done (and the initial file check if done)
		for( int a = 0; a <= xferNo; ++a )
		{
			stats[10][GOOD_BYTE_COUNT] = stats[10][GOOD_BYTE_COUNT] + stats[a][GOOD_BYTE_COUNT];
			stats[10][RECOV_BYTE_COUNT] = stats[10][RECOV_BYTE_COUNT] + stats[a][RECOV_BYTE_COUNT];
			stats[10][BAD_BYTE_COUNT] = stats[10][BAD_BYTE_COUNT] + stats[a][BAD_BYTE_COUNT];
			stats[10][TOT_BYTE_COUNT] = stats[10][TOT_BYTE_COUNT] + stats[a][TOT_BYTE_COUNT];
			stats[10][ELAPSED_SECS] = stats[10][ELAPSED_SECS] + stats[a][ELAPSED_SECS];
			if(a > 0)
				stats[10][RATE_BYTES] = (stats[10][RATE_BYTES] * (a - 1) + stats[a][RATE_BYTES]) / a;
		}

		emit uploadProgress(
			(int)pct,
			"Summary",
			((double)stats[10][GOOD_BYTE_COUNT] / 1048576.0),
			((double)stats[10][RECOV_BYTE_COUNT] / 1048576.0),
			((double)stats[10][BAD_BYTE_COUNT] / 1048576.0),
			((double)stats[10][TOT_BYTE_COUNT] / 1048576.0),
			((double)stats[10][RATE_BYTES] / 1024.0),
			(int)stats[10][ELAPSED_SECS],
			int(-1));
	}

	// Display final transfer status to user
	if(xferGetGoodByteCount(&goodBlocks) - padFileBytes != sourceFile->size)
	{
		// File transfer not completed due to user abort or if Transfer Error Recovery option was disabled by
		// the user in Settings
		if(escape)
			printf("  ***USER ABORT***    %sONLY %2.2f%% OF THE FILE HAS BEEN SUCCESSFULLY TRANSFERRED\n", (pct < 10 ? " " : ""), pct);
		else
			printf("  ***WARNING***       %sONLY %2.2f%% OF THE FILE HAS BEEN SUCCESSFULLY TRANSFERRED\n", (pct < 10 ? " " : ""), pct);
	}
	else
	{
		// File transfer already done or has completed successfully
		if(xferNo == 0)
			printf("  THE FILE HAS ALREADY BEEN SUCCESSFULLY TRANSFERRED\n");
		else
			printf("  THE FILE WAS SUCCESSFULLY TRANSFERRED\n");
	}
	printf("  -----------------------------------------------------------------------------\n\n");

	// Delete Good Block list memory
	deleteBlockList(&goodBlocks);

	// Return true if successful or false if the user pressed ESCAPE to abort
	return !escape;
}

// Transfer a single file from host to Humax via USB
// Inputs
// - sourceFile : source (Host) file
// - targetFile : target (Humax) file
// Returns
// - true       : if operation completed successully
// - false      : if file transfer failed with serious USB error or user aborted with the ESCAPE key
bool HumaxController::xferFileHostToDevice(Entry *sourceFile, Entry *targetFile)
{
	// Open the source host file for reading
	FILE * inFile = NULL;
///	if(fopen_s(&inFile, sourceFile->path, "rb") != 0)
	if( (inFile = fopen(sourceFile->path, "rb")) == NULL )
	{
		emit transferError(QString("Failed to open source file '%1'!").arg(sourceFile->path));
		return false;
	}

	// Delete the target file on the Humax if it exists
	humaxDeleteFile(targetFile->path);

	// Display transfer stats header to the user
	QString name1 = getFileName(sourceFile->path);
	QString name2 = getFileName(targetFile->path);
	bool nameChange = (name1 != name2);
	printf("  Transfer %s'%s'\n", (nameChange ? " " : ""), name1.toStdString().data());
	if(nameChange)
		printf("  Rename to '%s'\n", name2.toStdString().data());
	printf("  From Host to Humax, Size %.3f MB\n", (double)sourceFile->size / 1048576.0);

	// Initiate the "M2" file transfer via USB
	int xferCBWTag = incCBWTag();
	if(!humaxXferBeginHostToDev(xferCBWTag, targetFile->path, targetFile->size))
	{
		emit transferError(QString("Failed to begin file transfer for '%1'!").arg(sourceFile->path));
		fclose(inFile);
		return false;
	}

	// Get transfer start time
	time_t startTime;
	time(&startTime);
	time_t nextTime = startTime + 1;

	// Loop for each data block until the source file has been transferred
	char data[0x10000];
	bool ok = true;
	bool escape = false;
	qint64 remainingByteCount = sourceFile->size;
	while(ok && !escape && remainingByteCount > 0)
	{
		// Determine size of data block to send based on the "Host-Humax Xfer Block Size" parameter in Settings
		unsigned int thisByteCount = downloadBlockSize * 256;

		// If we are at the end of the source host file then adjust the data block size accordingly
		if(thisByteCount > remainingByteCount)
			thisByteCount = (int)remainingByteCount;

		// Read data from source host file into buffer
		unsigned int cbRet = fread(data, 1, thisByteCount, inFile);
		if(cbRet != thisByteCount)
		{
			emit transferError(QString("Failed to read data from file '%1'!").arg(sourceFile->path));
			ok = false;
			continue;
		}

		// Send data block via USB to the Humax
		ok = humaxXferNextHostToDev(xferCBWTag, data, thisByteCount);
		if(!ok)
		{
			emit transferError(QString("Failed to send data to device!"));
			ok = false;
			continue;
		}

		// Update remaining transfer byte count
		remainingByteCount = remainingByteCount - thisByteCount;

		// Display status to user after each second elapses
		time_t curTime;
		time(&curTime);
		if(curTime >= nextTime)
		{
			qint64 totByteCount = sourceFile->size - remainingByteCount;
			int elapsedSecs = (int)(curTime - startTime);
			if(elapsedSecs == 0)
				elapsedSecs = 1;
			int rateBytes = (int)(totByteCount / elapsedSecs);
			int remainSecs = (int)(((double)sourceFile->size - totByteCount) / ((double)totByteCount / elapsedSecs));

			emit downloadProgress((int)((double)totByteCount / sourceFile->size * 100.0),
					((double)totByteCount / 1048576.0),
					((double)remainingByteCount / 1048576.0),
					((double)rateBytes / 1024.0),
					elapsedSecs,
					remainSecs);

			nextTime = curTime + 1;
		}

		// Check if ESCAPE key pressed to abort by user
		if(!escape)
			escape = mCancelled;
	}

	// End the file transfer via USB
	if(!humaxXferEnd(xferCBWTag, targetFile->path))
	{
		emit transferError(QString("Failed to end file transfer for '%1'!").arg(sourceFile->path));
		fclose(inFile);
		return false;
	}

	// Close the host source input file
	fclose(inFile);

	// Display final transfer stats to user
	time_t curTime;
	time(&curTime);
	qint64 totByteCount = sourceFile->size - remainingByteCount;
	int elapsedSecs = (int)(curTime - startTime);
	if(elapsedSecs == 0)
		elapsedSecs = 1;
	int rateBytes = (int)(totByteCount / elapsedSecs);
	int remainSecs = (int)(((double)sourceFile->size - totByteCount) / ((double)totByteCount / elapsedSecs));

	emit downloadProgress((int)((double)totByteCount / sourceFile->size * 100.0),
			((double)totByteCount / 1048576.0),
			((double)remainingByteCount / 1048576.0),
			((double)rateBytes / 1024.0),
			elapsedSecs,
			remainSecs);

	double pct = ((double)totByteCount / sourceFile->size * 100.0);

	// Display final transfer status to user
	if(remainingByteCount != 0)
	{
		// If user aborted the transfer or a serious error occurred then display total percentage of file actually
		// transferred
		if(escape)
			printf("  ***USER ABORT***       %sONLY %2.2f%% OF THE FILE TRANSFERRED\n", (pct < 10 ? " " : ""), pct);
		else
			printf("  ***XFER ERROR***       %sONLY %2.2f%% OF THE FILE TRANSFERRED\n", (pct < 10 ? " " : ""), pct);

		// We delete the partial file on the Humax here because if a user tries to display, say, a partial JPG, or play
		// a partial MP3 file on their unit it could possibly hang the device. So play safe just in case.
		humaxDeleteFile(targetFile->path);
	}
	else
	{
		// File transfer successfully transferred
		printf("  THE FILE WAS SUCCESSFULLY TRANSFERRED\n");
	}

	// Return true if success or false if the user pressed the ESCAPE key to abort the transfer
	return !escape;
}

//////////////////////////////////////////////////////////////////////////////////////
///// HIGH_LEVEL USB DEVICE FUNCTIONS                                            /////
//////////////////////////////////////////////////////////////////////////////////////

// Sets the new current device disk and directory and reads the directory index. Will return the actual number
// of directory entries present although it will only return a maximum of 896 entries due to USB data block size
// limitations.
// Inputs
// - dir : new directory path to change to on the device
// Returns
// - dir : linked-list of child directory entries
// -  -1 : an error occurred
// -   0 : directory contains no entries
// -  >0 : number of entries in the new directory

// Returns the actual no directory entries although only the first 896 entries will be read
// Returns -1 if an error occurred
int HumaxController::setDeviceDirectory(Entry *dir)
{
	// Delete any existing entry tree
	deleteEntryTree(dir);

	// Set disk partition no
	if(!humaxSetDiskNo(dir->path[0]))
	{
		emit transferError(QString("Failed to set device disk no %1!").arg(dir->path[0]));
		return -1;
	}

	// Set current directory
	if(!humaxSetDirectory(dir->path))
	{
		emit transferError(QString("Failed to set device directory for '%1'!").arg(dir->path));
		return -1;
	}

	// Get number of directory entries
	int noEntries = humaxGetNoDirEntries();
	if(noEntries == -1)
	{
		emit transferError(QString("Failed to get file count for directory '%1'!").arg(dir->path));
		return -1;
	}
	if(noEntries == 0)
		return 0;

	// Limit the number of directory entries to read from the Humax to 896.
	// Reading any more than this will hang the Humax because it exceeds the maximum USB data block size!
	int noEntriesLimited = noEntries;
	if(noEntriesLimited > HUMAX_MAX_DIR_ENTRIES)
		noEntriesLimited = HUMAX_MAX_DIR_ENTRIES;

	// Allocate memory
	char * data = new char[noEntriesLimited * 0x48];
	bool * sizeChanged = new bool[noEntriesLimited];

	// If reading the Video disk then do a dual read so we can get the recording status of files.
	// This works by simply reading the directory contents, pausing for 800ms, reading again a second
	// time then comparing the file sizes between the two. Any that are different means the file is
	// currently being recorded.
	bool doDualRead = (strcmp(dir->path, "2:/") == 0);

	// Read the directory data (twice with 800ms pause if video directory)
	bool ok = true;
	if(humaxGetDirEntries(noEntriesLimited, doDualRead, data, sizeChanged))
	{
		// Create directory entries from the returned data
		buildDeviceEntryList(data, sizeChanged, noEntriesLimited, dir);
	}
	else
	{
		emit transferError(QString("Failed to retrieve directory contents for '%1'!").arg(dir->path));
		ok = false;
	}

	// Free data structures
	free(data);
	free(sizeChanged);

	// Return status
	return (ok ? noEntries : -1);
}

// Builds a complete device entry tree in memory from the current 'dir' directory downwords
// Inputs
// - dir   : top-level directory for which to build the tree
// - level : always set to 0
// - stats : input value ignored, values set on exit from function
// Returns
// - dir   : hierarchical linked-list tree of child and sibling entries containing the tree structure
// - stats : [0] total entries [1] total entries size [2] total dirs [3] total dirs size [4] total files [5] total files size
int HumaxController::buildDeviceEntryTree(Entry *dir, int level, qint64 *stats)
{
	// If top level then initialise stats
	if(level == 0)
	{
		// Set current disk no
		if(!humaxSetDiskNo(dir->path[0]))
		{
			emit transferError(QString("Failed to set disk partition %1!").arg(dir->path[0]));
			return -1;
		}

		// Clear stats
		for(int a = 0; a < 6; a++)
			stats[a] = 0;
	}

	// Set current directory
	if(!humaxSetDirectory(dir->path))
	{
		emit transferError(QString("Failed to set directory for '%1'!").arg(dir->path));
		return -1;
	}

	// Get number of directory entries in current directory
	int noEntries = humaxGetNoDirEntries();
	if(noEntries == -1)
	{
		emit transferError(QString("Failed to get file count for directory '%1'!").arg(dir->path));
		return -1;
	}
	if(noEntries == 0)
		return 0;

	// If more than 896 entries then limit to 896 to prevent blowing max Humax USB data block size
	if(noEntries > HUMAX_MAX_DIR_ENTRIES)
		noEntries = HUMAX_MAX_DIR_ENTRIES;

	// Get directory entry data
	char * data = new char[noEntries * 0x48];
	bool * sizeChanged = new bool[noEntries];

	// Do dual read if video files directory
	bool doDualRead = (strcmp(dir->path, "2:/") == 0);

	// Read directory entries data
	if(humaxGetDirEntries(noEntries, doDualRead, data, sizeChanged))
	{
		// Create directory entries from the returned data
		buildDeviceEntryList(data, sizeChanged, noEntries, dir);
	}
	else
	{
		emit transferError(QString("Failed to retrieve directory contents for directory '%1'!").arg(dir->path));
		free(data);
		free(sizeChanged);
		return -1;
	}

	// Free data structures
	free(data);
	free(sizeChanged);

	// Loop for each entry in the current directory
	Entry * e = dir->firstChild;
	while(e != NULL)
	{
		// Update stats
		stats[ENTRIES_COUNT]++;
		stats[ENTRIES_SIZE] = stats[ENTRIES_SIZE] + e->size;

		// Check if file or directory
		if(isDirectory(e))
		{
			// Current entry is a sub-directory so update stats and recurse into it
			stats[DIRS_COUNT]++;
			stats[DIRS_SIZE] = stats[DIRS_SIZE] + e->size;
			if(buildDeviceEntryTree(e, level + 1, stats) == -1)
				return -1;
		}
		else
		{
			// Current entry is a file so just update stats
			stats[FILES_COUNT]++;
			stats[FILES_SIZE] = stats[FILES_SIZE] + e->size;
		}

		// Do next entry
		e = e->sibling;
	}

	// Return total number of file + directory entries read
	return (int)stats[ENTRIES_COUNT];
}

// Checks the first 1KB of a Host PC file with the same file on the Humax (reads the first 1KB of the file over USB)
// and returns a status indicator after comparing the two files.
// Inputs
// - sourceFile     : source file (host or device) to check/compare
// - targetFile     : target file (host or device) to check/compare
// Returns
// - FILE_NOT_FOUND : source file not found at the target
// - FILE_EQUAL     : source file is the same as the target (first 1KB of data is the same)
// - FILE_DIFFERENT : source file is different from the target (first 1KB of data is
//                    different or target file is bigger than the source file)
int HumaxController::checkFile(Entry *sourceFile, Entry *targetDir)
{
	// Determine what host and target are
	char hostData[1024];
	char deviceData[1024];
	char name[MAX_PATH];
	Entry * hostFile;
	Entry * deviceFile;
	sprintf(name, "%s", getFileName(sourceFile->path).toStdString().data());

	if(sourceFile->path[0] == '1' || sourceFile->path[0] == '2')
	{
		// Source is Humax file and target is host file. Check if source exists on the target
		getValidHostFilePath(name, strlen(name));
		deviceFile = sourceFile;
		hostFile = targetDir->firstChild;

		QFileInfo targetInfo(targetDir->path);

		// Source file not found on target so no overwrite issues
		if( !targetInfo.exists() )
			return FILE_NOT_FOUND;

		// Target file exists and is larger in size than the source file so assume these files are different
		// for overwrite purposes
		if( targetInfo.size() > deviceFile->size )
			return FILE_DIFFERENT;
	}
	else
	{
		// Source is host file and target is Humax file. Check if source exists on the target
		getValidDeviceFilePath(name, strlen(name));

		hostFile = sourceFile;
		deviceFile = targetDir->firstChild;
		bool found = false;
		while(!found && deviceFile != NULL)
		{
			QString name2 = getFileName(deviceFile->path);
			found = (QString(name) == name2);
			if(!found)
				deviceFile = deviceFile->sibling;
		}

		// Source file not found on target so no overwrite issues
		if(deviceFile == NULL)
		{
			return FILE_NOT_FOUND;
		}

		// Target file exists and is larger in size than the source file so assume these files are different
		if(deviceFile->size > hostFile->size)
			return FILE_DIFFERENT;
	}

	// Read the first 1KB of data from the host file and if less than 1KB data available then assume the
	// source and target files are different for overwrite purposes
	FILE *inFile = NULL;
///	if(fopen_s(&inFile, hostFile->path, "rb") != 0)
	if( (inFile = fopen(hostFile->path, "rb")) == NULL )
		return FILE_DIFFERENT;
	int hostFileSize = fread(hostData, 1, 1024, inFile);
	fclose(inFile);
	if(hostFileSize < 1024)
		return FILE_DIFFERENT;

	// Read first 1KB of data from the Humax file via USB and if less than 1KB data available then assume the
	// source and target files are different for overwrite purposes
	int xferCBWTag = incCBWTag();
	if(!humaxXferBeginDevToHost(xferCBWTag, deviceFile->path))
		return FILE_DIFFERENT;
	int deviceFileSize = humaxXferNextDevToHost(xferCBWTag, deviceData, 1024);
	if(!humaxXferEnd(xferCBWTag, deviceFile->path))
		return FILE_DIFFERENT;
	if(deviceFileSize < 1024)
		return FILE_DIFFERENT;

	// Compare the first 1KB of data between source and target files. If data differs then assume files are
	// different for overwrite purposes
	for(int a = 0; a < 1024; a++)
		if(hostData[a] != deviceData[a])
			return FILE_DIFFERENT;

	// Got this far so assumption is the source and target are the same files
	return FILE_EQUAL;
}

// Gets the file size and recording status information for a given Humax or host file
// Inputs
// - file  : host or device file to get file size info for
// Returns
// - file  : host or device file is updated with new file size and 'isRecording' flag status (for Humax video files)
// - true  : operation succeeded
// - false : operation failed
bool HumaxController::getFileInfo(Entry *file)
{
	bool ok = true;
	file->size = 0;
	file->isRecording = false;
	if(file->path[0] == '1' || file->path[0] == '2')
	{
		// Get info for a device file
		Entry parentDir;
		initEntry(&parentDir, NULL, 0, 0);
		getFilePath(file->path, file->nameLen, NULL, false, parentDir.path);

		if(setDeviceDirectory(&parentDir) == -1)
		{
			deleteEntryTree(&parentDir);
			emit transferError(QString("Failed to get device file info for '%1'").arg(file->path));
			return false;
		}
		Entry * e = parentDir.firstChild;
		while(e != NULL && strcmp(e->path, file->path) != 0)
			e = e->sibling;
		ok = (e != NULL);
		if(ok)
		{
			file->size = e->size;
			file->isRecording = e->isRecording;
		}
		deleteEntryTree(&parentDir);
	}
	else
	{
		// Get info for a host file
		QFileInfo info(file->path);
		if( info.exists() )
		{
			file->isRecording = false;
			file->size = info.size();
		}
	}
	return ok;
}

// Displays a list of files and directories in a given directory returning the total number of entries displayed
// Inputs
// - dir            : directory to get entries for
// - displayDirs    : true if directories are to be displayed
// - displayFiles   : true if files are to be displayed
// - displayLinks   : true if '/', '.' and '..' directory links are to be displayed
// - displaySummary : true if header/footer info to be displayed
// - displayIndexes : true if selection indexes are to be displayed
// - stats          : array of 6 integers, value ignored on entry
// Returns
// - dir            : linked-list file/directory listing of entries
// - stats          : [0] tot entries [1] tot entries size [2] tot dirs [3] tot dirs size [4] tot files [5] tot files size
// - int            : total number of entries displayed (same as stats[0])
int HumaxController::displayEntryList(Entry *dir, bool displayDirs, bool displayFiles, bool displayLinks, bool displaySummary, bool displayIndexes, qint64 *stats)
{
	// Initialise counters
	qint64 diskSpace[2];
	for(int a = 0; a < 6; a++)
		stats[a] = 0;
	dir->index = -1;

	// Check if summary display requested
	if(displaySummary)
	{
		// Get free disk space
		humaxGetDiskSpace(diskSpace);
	}

	// Display root [/] current [.] and parent [..] directory links but only if not top-level
	if(displayLinks && !isRootDirectory(dir))
	{
		stats[ENTRIES_COUNT] = stats[DIRS_COUNT] = 3;
	}

	// Loop for all files in the directory
	Entry * e = dir->firstChild;
	while(e != NULL)
	{
		e->index = -1;
		bool isDir = isDirectory(e);
		if((isDir && displayDirs) || (!isDir && displayFiles))
		{
			// Got valid entry for display so update stats
			stats[ENTRIES_COUNT]++;
			stats[ENTRIES_SIZE] = stats[ENTRIES_SIZE] + e->size;
			e->index = (int)stats[ENTRIES_COUNT];
			if(isDir)
			{
				stats[DIRS_COUNT]++;
				stats[DIRS_SIZE] = stats[DIRS_SIZE] + e->size;
			}
			else
			{
				stats[FILES_COUNT]++;
				stats[FILES_SIZE] = stats[FILES_SIZE] + e->size;
			}

			// Get name of file or directory and display it
			QString name = getFileName(e->path);
		}

		// Do next entry
		e = e->sibling;
	}

	// Display summary if requested
	if(displaySummary)
	{
		printf("  %s%s------------------------------------------------------%s%s\n", (displayIndexes ? "------ " : ""), (displayDirs && !displayFiles ? "" : "------------  "), (displayIndexes ? "" : "-------"), (displayDirs && !displayFiles ? "--------------" : ""));
		if(displayDirs && !displayFiles)
			printf("  %sTotal %d dir%s (%.2f MB free)\n\n", (displayIndexes ? "       " : ""), (int)stats[DIRS_COUNT], (stats[DIRS_COUNT] == 1 ? "" : "s"), (double)diskSpace[0] / 1048576.0);
		else
			printf("  %s%9.2f MB  Total %d dir%s, %d file%s (%.2f MB free)\n\n", (displayIndexes ? "       " : ""), (double)stats[FILES_SIZE] / 1048576.0, (int)stats[DIRS_COUNT], (stats[DIRS_COUNT] == 1 ? "" : "s"), (int)stats[FILES_COUNT], (stats[FILES_COUNT] == 1 ? "" : "s"), (double)diskSpace[0] / 1048576.0);
	}

	// Return number of entries displayed
	return (int)stats[ENTRIES_COUNT];
}

// Builds a device directory entry list from the data returned from a call to the humaxGetDirEntries() function
// Inputs
// - data        : raw data read from Humax device via the humaxGetDirEntries() function
// - sizeChanged : array of boolean flags to indicate if the file has changed in size (values set on exit)
// - noEntries   : number of entries to process from the data buffer
// - dir         : directory entry in which to store the new child entries
// Returns
// - sizeChanged : array of boolean flags to indicate if the file has changed in size
// - dir         : contains linked-list of new child entries
void HumaxController::buildDeviceEntryList(char * data, bool * sizeChanged, int noEntries, Entry * dir)
{
	// If no entries then just exit
	if(noEntries == 0)
		return;

	// Build the entry list
	for(int a = 0; a < noEntries; a++)
	{
		// Read current entry (0x48 or 72 bytes per entry)
		char path[MAX_PATH];
		int nameLen = 0;
		while(nameLen < 0x3c && data[a * 0x48 + nameLen] != 0)
			nameLen++;
		int pathLen = strlen(dir->path);
		bool isDir = ((data[a * 0x48 + 0x3c] & 0x01) != 0);
		qint64 size = (((qint64)((unsigned char)data[a * 0x48 + 0x40]) * 1000000000) + (unsigned char)data[a * 0x48 + 0x44] + ((unsigned char)data[a * 0x48 + 0x45] * 256) + ((unsigned char)data[a * 0x48 + 0x46] * 65536) + ((unsigned char)data[a * 0x48 + 0x47] * 16777216));
		strcpy(path, dir->path);
///		strcpy_s(path, dir->path);
		memcpy(path + pathLen, data + (a * 0x48), nameLen);
		path[pathLen + nameLen] = 0;
		if(isDir)
		{
			path[pathLen + nameLen] = '/';
			path[pathLen + nameLen + 1] = 0;
		}
		Entry * newEntry = createEntry(path, (isDir ? 0 : nameLen), size);
		newEntry->isRecording = sizeChanged[a];

		// Insertion sort the entry into the current list
		Entry * prevEntry = NULL;
		Entry * curEntry = dir->firstChild;
		if(curEntry == NULL)
		{
			// New entry is first child of parent directory
			dir->firstChild = newEntry;
		}
		else
		{
			// Get name of new entry
			QString name = getFileName(newEntry->path);
			name = name.toLower();

			// Insertion sort entry into list
			bool done = false;
			while(!done && curEntry != NULL)
			{
				// Get entry name
				QString name2 = getFileName(curEntry->path);
				name2 = name2.toLower();
				bool isDir2 = isDirectory(curEntry);

				// Check if this is the correct place
				done = (isDir && !isDir2) || (((isDir && isDir2) || (!isDir && !isDir2)) && name < name2);
				if(!done)
				{
					// Not correct place so check next entry
					prevEntry = curEntry;
					curEntry = curEntry->sibling;
				}
			}

			// Link entry into existing list
			Entry * temp = NULL;
			if(prevEntry == NULL)
			{
				// New entry is first in the list
				temp = dir->firstChild;
				dir->firstChild = newEntry;
			}
			else
			{
				// New entry is not first in list
				temp = prevEntry->sibling;
				prevEntry->sibling = newEntry;
			}
			newEntry->sibling = temp;
		}
	}
}

/*!
 * Extracts just the filename of a file or directory.
 * \param filePath	input host or device filespec.
 * \return Filename string.
 */
QString HumaxController::getFileName(const QString &filePath)
{
	QString retVal = filePath;
	if( retVal.right(1) == "/" )
		retVal.resize(retVal.length() - 1);

	int pos = retVal.lastIndexOf('/');
	if( pos >= 0 )
	{
		return retVal.mid(pos + 1);
	}
	return QString();
}

// Checks the supplied file size against the Good Block list to determine if any bad blocks still exist and so
// whether a transfer has been fully completed
// Inputs
// - fileSize : size of file to check
// - blocks   : Good Blocks list to scan against
// Returns
// - true     : no Bad Blocks found so file transfer is complete
// - false    : Bad Blocks still found so transfer is incomplete
bool HumaxController::isXferComplete(qint64 fileSize, Block *blocks)
{
	// If bad data at start of block list then return false
	Block * b = blocks;
	if(b->begByteNo != 0)
		return false;

	// Loop to find any gaps in the Good Block list. If any found then return false.
	while(b != NULL)
	{
		if((b->nextBlock == NULL && b->endByteNo + 1 != fileSize) || (b->nextBlock != NULL && b->endByteNo + 1 != b->nextBlock->begByteNo))
			return false;
		b = b->nextBlock;
	}

	// Got this far so everything passed ok and we have a complete Good Block list which matches the supplied file size
	// so return true
	return true;
}

// Returns the total number of bytes required to be read again via USB to error-correct a file (for recovery transfers)
// Inputs
// - blocks : Good Blocks list to scan
// Returns
// - int    : total number of bytes required to be read again for recovert transfer
qint64 HumaxController::xferGetRecoveryByteCount(Block *blocks)
{
	Block * b = blocks;
	while( b->nextBlock != NULL )
		b = b->nextBlock;
	return b->begByteNo;
}

// Returns the total number of good bytes currently received for a file
// Inputs
// - blocks : Good Blocks list to scan
// Returns
// - int    : total number of good bytes received so far
qint64 HumaxController::xferGetGoodByteCount(Block * blocks)
{
	Block * b = blocks;
	qint64 goodByteCount = 0;
	while(b != NULL)
	{
		goodByteCount = goodByteCount + b->endByteNo - b->begByteNo + 1;
		b = b->nextBlock;
	}
	return goodByteCount;
}

// Return the current device directory
// Returns
// - Entry : Returns the current device directory entry for the current disk number
Entry *HumaxController::getCurDeviceDir()
{
	return &deviceCurDir[deviceCurDiskNo - '1'];
}

// Check if an Entry is a directory
// Inputs
// - e     : Entry to check
// Returns
// - true  : entry is a directory (i.e. filepath ends in '/' for Humax dirs, '\' for host dirs)
// - false : entry is not a directory
bool HumaxController::isDirectory(Entry * e)
{
	return (e->nameLen == 0);
}

// Checks if an Entry is a root directory
// Inputs
// - e     : Entry to check
// Returns
// - true  : entry is root directory (e.g. "1:/" or "2:/" for Humax, "c:\" or "d:\" for host)
// - false : entry is not root directory
bool HumaxController::isRootDirectory(Entry * e)
{
	return (strlen(e->path) == 3 && e->nameLen == 0 && e->path[2] == '/' || e->path[2] == '\\');
}

// Gets the directory path of an existing filespec and appends the new name if supplied
// If forceParent is false, returns the directory path of a file or just returns the existing path of a directory
// If forceParent is true, always returns the parent directory path of a file or directory filespec
// Inputs
// - path        : host or device filespec
// - nameLen     : length of filename in path above
// - newName     : new filename to use
// - forceParent : true, force extraction of the parent dirspec for dirs
//               : false, return the same filespec for dirs
// - filepath    : return buffer, value ignored on entry
// Returns
// - filepath    : buffer to receive conjoined path extraction and new filename (if specified)
// - int         : length of returned filepath
int HumaxController::getFilePath(char *path, int nameLen, char *newName, bool forceParent, char *filepath)
{
	int pathLen = strlen(path);
	if(nameLen > 0)
	{
		// Name present in path so strip off name and leave parent dirpath
		memcpy(filepath, path, pathLen - nameLen);
		filepath[pathLen - nameLen] = 0;
	}
	else if(forceParent && pathLen > 3)
	{
		// No name present and we want the parent of the dirpath so get parent dirpath
		pathLen--;
		pathLen--;
		while(pathLen > 2 && path[pathLen] != '/' && path[pathLen] != '\\')
			pathLen--;
		memcpy(filepath, path, pathLen + 1);
		filepath[pathLen + 1] = 0;
	}
	else
	{
		// Already a dirpath and parent dirpath not requested so just return current dirpath as-is
		strncpy(filepath, path, MAX_PATH);
///		strcpy_s(filepath, MAX_PATH, path);
	}

	// If name supplied then append new filename
	if(newName != NULL)
		strncpy(filepath, newName, MAX_PATH);
///		strcat_s(filepath, MAX_PATH, newName);

	// Return length of new path
	return strlen(filepath);
}

////////////////////////////////////////////////////////////////////////////////
/*! \brief Returns a valid device filepath by truncating the filename.
 *
 * Returns a valid device filepath by truncating the filename to a maximum of
 * 59 characters. If the filetype after the '.' character in the filename is 4
 * or less characters then preserves that filetype.
 * \param filepath	Input filepath, Output modified valid filepath.
 * \param nameLen	Length of filename in filepath.
 * \return Length of returned new filename length part of.
 */
////////////////////////////////////////////////////////////////////////////////
int HumaxController::getValidDeviceFilePath(char *filepath, int nameLen)
{
	// Filename present so validate and truncate it if necessary, preserving filetypes of 4 chars or less
	if(nameLen > HUMAX_MAX_FILENAME_LEN)
	{
		// Locate filetype no larger than 4 chars in length
		char name[MAX_PATH];
		char path[MAX_PATH];
		sprintf(name,"%s",getFileName(filepath).toStdString().data());
		getFilePath(filepath, nameLen, NULL, false, path);
		int p = nameLen - 1;
		int ftLen = 0;
		while(p >= 0 && ftLen < 5 && name[p] != '.')
		{
			p--;
			ftLen++;
		}
		p--;
		ftLen++;
		if(ftLen < 6)
			strncpy(name + HUMAX_MAX_FILENAME_LEN - ftLen, name + nameLen - ftLen, MAX_PATH);
///			strcpy_s(name + HUMAX_MAX_FILENAME_LEN - ftLen, MAX_PATH, name + nameLen - ftLen);
		name[HUMAX_MAX_FILENAME_LEN] = 0;
		strcat(path, name);
///		strcat_s(path, name);
		strncpy(filepath, path, MAX_PATH);
///		strcpy_s(filepath, MAX_PATH, path);
		nameLen = HUMAX_MAX_FILENAME_LEN;
	}
	return nameLen;
}

////////////////////////////////////////////////////////////////////////////////
/*! \brief Returns a valid host filepath.
 *
 * Returns a valid host filepath by replacing any invalid characters in
 * the filename with the underscore '_' character.
 * \param filepath	Input filepath, Output modified valid filepath.
 * \param nameLen	Length of filename in filepath.
 * \return Length of filename part only.
 */
////////////////////////////////////////////////////////////////////////////////
int HumaxController::getValidHostFilePath(char * filepath, int nameLen)
{
	if(nameLen > 0)
	{
		const QString INVALID_CHARS("\\/:*?\"<>|");
		int pathLen = strlen(filepath);
		for( int a = pathLen - nameLen; a < pathLen; ++a )
			if(filepath[a] < 32 || filepath[a] > 126 || INVALID_CHARS.indexOf(filepath[a]) != -1)
				filepath[a] = '_';
	}
	return nameLen;
}
////////////////////////////////////////////////////////////////////////////////
