////////////////////////////////////////////////////////////////////////////////
// \namespace Humax 
// \file HumaxGlobals.cpp
//
// \version 1.00
// \date 14-07-2006
// \author Andrew Smith <andrew.smith@freenet.co.uk>
//
// \par Copyright:
// See COPYING file that comes with this distribution.
//
// \brief Contains any global utility functions.
//
// A namespace to contain any utility functions needed within this application.
//
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
///// CHANGE HISTORY                                                       /////
////////////////////////////////////////////////////////////////////////////////
/*****

Date:     Ver: Description:
-------- ---- ------------------------------------------------------------------
14-07-06 1.00 Initial release.

*****/
#include "HumaxGlobals.h"

#include <QStringList>

////////////////////////////////////////////////////////////////////////////////
/* \brief Converts the size to a string.
 *
 * Converts the 'size' in bytes to a string with the units shown,
 * Eg bytes, KB, or MB.
 * \param size			The size in bytes to be converted.
 * \param fieldWidth	The width of the number part of the string.
 * \param prec			The floating point precision to be used in the string.
 * \return The size converted to a string.
 */
////////////////////////////////////////////////////////////////////////////////
const QString Humax::SizeAsString(	const qint64 &size,
									const int fieldWidth,
									const int prec)
{
	QStringList sizes;
	sizes << "bytes" << "KB" << "MB" << "GB" << "TB";

	double dSize(size);
	while( sizes.count() > 1 )
	{
		if( dSize > 1024.0 )
		{
			dSize /= double(1024);
			sizes.removeFirst();
		}
		else
		{
			break;
		}
	}
/*	if ($size < '10') $size = intval($size*100)/100;
	else if ($size < '100') $size = intval($size*10)/10;
	else $size = intval($size);
*/	return QString("%L1 %2").arg(dSize, fieldWidth, 'f', prec).arg(sizes[0]);
}

////////////////////////////////////////////////////////////////////////////////
