////////////////////////////////////////////////////////////////////////////////
// \class QDiskBar QDiskBar.cpp
//
// \version 1.00
// \date 04-07-2006
// \author Andrew Smith <andrew.smith@freenet.co.uk>
//
// \par Copyright:
// See COPYING file that comes with this distribution.
//
// \brief Disk Information Bar Widget Class.
//
// A Qt Widget class to graphically display the space used on a disk partition.
//
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
///// CHANGE HISTORY                                                       /////
////////////////////////////////////////////////////////////////////////////////
/*****

Date:     Ver: Description:
-------- ---- ------------------------------------------------------------------
04-07-06 1.00 Initial release.

*****/
#include "QDiskBar.h"

#include <QPainter>
#include <QPaintEvent>

////////////////////////////////////////////////////////////////////////////////
/*! \brief Constructor.
 *
 * Construct and initialise the data in the Disk Bar Class.
 */
////////////////////////////////////////////////////////////////////////////////
QDiskBar::QDiskBar(QWidget *parent)
: QFrame(parent)
, mFreeSpace(0)
, mTotalSpace(0)
{
}

////////////////////////////////////////////////////////////////////////////////
/*! \brief Paint the Disk Bar on the screen.
 *
 * Pain the Disk Bar on the dialog when a paint request event is made.
 * \param event	The paint event request.
 */
////////////////////////////////////////////////////////////////////////////////
void QDiskBar::paintEvent(QPaintEvent *event)
{
	QFrame::paintEvent(event);

	if( mTotalSpace > 0 )
	{
		QPainter painter(this);

		const int border = 2;
		QRect drawRect = event->rect().
			adjusted(border, border, -(border), -(border));

		painter.fillRect(drawRect, QColor(0,144,0));

		double usedPercent = double(mTotalSpace - mFreeSpace)
							/ double(mTotalSpace);
		int usedWidth = int(double(width() - 4) * usedPercent);

		painter.setPen(QColor(255,255,255));

		if( usedWidth > 0 )
		{
			drawRect.setWidth(usedWidth);
			painter.fillRect(drawRect, QColor(160,0,0));

			if( usedWidth > 32 )
			{
				painter.drawText(drawRect, Qt::AlignCenter,
					QString("%1%").arg(usedPercent * double(100), 0, 'f', 1));
			}
		}

		drawRect.setX(drawRect.x() + usedWidth);
		drawRect.setWidth((width() - 4) - usedWidth);

		if( drawRect.width() > 32 )
		{
			painter.drawText(drawRect, Qt::AlignCenter, QString("%1%").
				arg((double(100) - (usedPercent * double(100))), 0, 'f', 1));
		}

		painter.end();
	}
}

////////////////////////////////////////////////////////////////////////////////
